'''
pyssh

PySSH is a Python library for programmatically controlling ssh and scp.
'''

class PySSHError(Exception): pass
class CannotRun(PySSHError): pass


# Check Python version
import sys
ver = getattr(sys, 'version_info', (1,))
if ver[0]<2:
	raise CannotRun, 'Python 2.0 or higher is required. This version is %s.' % sys.version

# Check for pty
try:
	import pty
except ImportError:
	raise CannotRun, 'The pty module is required but cannot be found.'


import os, time, signal


DEFAULT_DELAY = 0.1
DEFAULT_BUFSIZE = 1024


_sshPath = None  # cache the path of the external ssh program
for path in ['/usr/bin/ssh', '/usr/local/bin/ssh']:
	if os.path.exists(path):
		_sshPath = path


def run(command, **specs):
	'''
	Runs, via ssh, a command specified by the specs and/or passed
	keywords. Required keys are:
		host
		command

	Optional keys are:
		user
		password
		debug
		sshPath
		delay       # for time.sleep()
		bufSize     # for os.read()

	Example:
		from pyssh import run
		run('date', host='host.org', user='user')

		# or if you need a password:
		run('date', host='host.org', user='user', password='password')

	See SSHController for a higher level interface.
	'''
	pid, fd = pty.fork()
	if pid==0:
		sshPath = specs.get('sshPath', _sshPath)
		if not sshPath:
			raise PySSHError, 'Cannot locate ssh external command.'
		args = [sshPath, specs['host'], command]
		os.execv(sshPath, args)
		sys.exit(0)
	else:
		delay = specs.get('delay', DEFAULT_DELAY) # @@ doc
		bufSize = specs.get('bufSize', DEFAULT_BUFSIZE) # @@ doc
		debug = specs.get('debug', None)
		if debug==1:
			debug = sys.stdout
		password = specs.get('password', None)
		if password:
			password += '\n'
		cmd = command + '\n'
		time.sleep(delay)
		if password:
			output = os.read(fd, bufSize)  # read the password prompt
			if debug: debug.write('>> child says: %r\n' % output)
			numBytes = os.write(fd, password)
			if debug:
				debug.write('>> child took %i password bytes\n' % numBytes)
				debug.write('>> sleeping %i secs\n' % delay)
			time.sleep(delay)  # being over cautious
		numBytes = os.write(fd, cmd)
		if debug: debug.write('>> child took %i command bytes\n' % numBytes)
		output = os.read(fd, bufSize)
		if debug: debug.write('>> child response is %r\n' % output)
		os.kill(pid, signal.SIGKILL)
		return output


class Result:
	'''
	Instances of Result are returned by SSHController's sendCommand().
	Attributes include:
		startTime
		endTime
		duration  # read only
		command
		output
		filteredOutput  # read only
		number # 1, 2, 3, ...
	'''

	def __getattr__(self, name):
		if name=='duration':
			return self.endTime - self.startTime
		elif name=='filteredOutput':
			return self._filteredOutput()
		else:
			raise AttributeError, '%s instance has no attribute %r' % (self.__class__.__name__, name)

	def pprint(self, out=None):
		if not out:
			out = sys.stdout
		out.write('%3i> %s\n' % (self.number, self.command))
		out.write('%0.2f secs\n' % self.duration)
		out.write('%r\n\n' % self.output)

	def _filteredOutput(self):
		output = self.output

		# get rid of \r's
		output = output.replace('\r', '')

		# get rid of the command which is often echoed back with the
		# output from the command
		output = output.replace(self.command+'\n', '')

		# strip preceding and trailing whitespace
		output = output.strip()

		self.filteredOutput = output  # cache the result
		return output


class SSHController:
	'''
	Unlike the run() function an ssh controller will:
		- locate the ssh and scp external commands in /usr/bin or /usr/local/bin if the paths are not specified
		- remember its state such as host, user, password, debug, etc.
		- return a Result object with additional info such as duration (in seconds), command, etc.
	'''

	def __init__(self, args={}, **keywords):
		self.host = None
		self.user = None
		self.password = None
		self.sshPath = None
		self.scpPath = None
		self.debug = 0
		self.commandNumber = 1

		self.__dict__.update(args)
		self.__dict__.update(keywords)

		if not self.sshPath:
			self.sshPath = _sshPath

		if not self.scpPath:
			for path in ['/usr/bin/scp', '/usr/local/bin/scp']:
				if os.path.exists(path):
					self.scpPath = path

	def run(self, command, **specs):
		self.__dict__.update(specs)
		result = Result()
		result.number = self.commandNumber
		self.commandNumber += 1
		result.command = command
		result.startTime = time.time()
		result.output = run(command, **self.__dict__)
		result.endTime = time.time()
		return result


def main(args=None):
	'''
	Runs ssh through python. Mostly useful for testing the module from
	the command line, easily allow you to specify arguments:
		python pyssh.py host=HOST user=USER command=COMMAND
	'''
	if args is None:
		args = sys.argv
	dict = {}
	for arg in args[1:]:
		key, value = arg.split('=')
		if key=='debug':
			try:
				value = int(value) # debug could be 0 or 1
			except ValueError:
				pass
		dict[key] = value
	results = run(**dict)
	print 'results ='
	print repr(results)


if __name__=='__main__':
	main()
